@@ -12,6 +12,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import collections
 import contextlib
 import errno
 import functools
@@ -27,7 +28,7 @@ import time
 from urllib import parse
 import uuid
 
-from amqp import exceptions as amqp_exec
+from amqp import exceptions as amqp_ex
 import kombu
 import kombu.connection
 import kombu.entity
@@ -35,6 +36,7 @@ import kombu.messaging
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import eventletutils
+from oslo_utils import netutils
 
 import oslo_messaging
 from oslo_messaging._drivers import amqp as rpc_amqp
@@ -45,6 +47,13 @@ from oslo_messaging._drivers import pool
 from oslo_messaging import _utils
 from oslo_messaging import exceptions
 
+
+# The QuorumMemConfig will hold the quorum queue memory configurations
+QuorumMemConfig = collections.namedtuple('QuorumMemConfig',
+                                         'delivery_limit'
+                                         ' max_memory_length'
+                                         ' max_memory_bytes')
+
 # NOTE(sileht): don't exist in py2 socket module
 TCP_USER_TIMEOUT = 18
 
@@ -84,21 +93,31 @@ rabbit_opts = [
                 'executable used does not support OpenSSL FIPS mode, '
                 'an exception will be raised.'),
     cfg.BoolOpt('heartbeat_in_pthread',
-                default=True,
-                help="Run the health check heartbeat thread "
+                default=False,
+                deprecated_for_removal=True,
+                deprecated_reason='The option is related to Eventlet which '
+                                  'will be removed. In addition this has '
+                                  'never worked as expected with services '
+                                  'using eventlet for core service framework.',
+                help="(DEPRECATED) It is recommend not to use this option "
+                     "anymore. Run the health check heartbeat thread "
                      "through a native python thread by default. If this "
                      "option is equal to False then the health check "
                      "heartbeat will inherit the execution model "
                      "from the parent process. For "
                      "example if the parent process has monkey patched the "
                      "stdlib by using eventlet/greenlet then the heartbeat "
-                     "will be run through a green thread.",
+                     "will be run through a green thread. "
+                     "This option should be set to True only for the "
+                     "wsgi services.",
                 ),
     cfg.FloatOpt('kombu_reconnect_delay',
                  default=1.0,
+                 min=0.0,
+                 max=amqpdriver.ACK_REQUEUE_EVERY_SECONDS_MAX * 0.9,
                  deprecated_group='DEFAULT',
-                 help='How long to wait before reconnecting in response to an '
-                      'AMQP consumer cancel notification.'),
+                 help='How long to wait (in seconds) before reconnecting in '
+                      'response to an AMQP consumer cancel notification.'),
     cfg.StrOpt('kombu_compression',
                help="EXPERIMENTAL: Possible values are: gzip, bz2. If not "
                     "set compression will not be used. This option may not "
@@ -153,16 +172,60 @@ rabbit_opts = [
                 'Raft consensus algorithm. It is available as of '
                 'RabbitMQ 3.8.0. If set this option will conflict with '
                 'the HA queues (``rabbit_ha_queues``) aka mirrored queues, '
-                'in other words the HA queues should be disabled, quorum '
-                'queues durable by default so the amqp_durable_queues '
-                'opion is ignored when this option enabled.'),
+                'in other words the HA queues should be disabled. '
+                'Quorum queues are also durable by default so the '
+                'amqp_durable_queues option is ignored when this option is '
+                'enabled.'),
+    cfg.BoolOpt('rabbit_transient_quorum_queue',
+                default=False,
+                help='Use quorum queues for transients queues in RabbitMQ. '
+                'Enabling this option will then make sure those queues are '
+                'also using quorum kind of rabbit queues, which are HA by '
+                'default.'),
+    cfg.IntOpt('rabbit_quorum_delivery_limit',
+               default=0,
+               help='Each time a message is redelivered to a consumer, '
+               'a counter is incremented. Once the redelivery count '
+               'exceeds the delivery limit the message gets dropped '
+               'or dead-lettered (if a DLX exchange has been configured) '
+               'Used only when rabbit_quorum_queue is enabled, '
+               'Default 0 which means dont set a limit.'),
+    cfg.IntOpt('rabbit_quorum_max_memory_length',
+               deprecated_name='rabbit_quroum_max_memory_length',
+               default=0,
+               help='By default all messages are maintained in memory '
+               'if a quorum queue grows in length it can put memory '
+               'pressure on a cluster. This option can limit the number '
+               'of messages in the quorum queue. '
+               'Used only when rabbit_quorum_queue is enabled, '
+               'Default 0 which means dont set a limit.'),
+    cfg.IntOpt('rabbit_quorum_max_memory_bytes',
+               deprecated_name='rabbit_quroum_max_memory_bytes',
+               default=0,
+               help='By default all messages are maintained in memory '
+               'if a quorum queue grows in length it can put memory '
+               'pressure on a cluster. This option can limit the number '
+               'of memory bytes used by the quorum queue. '
+               'Used only when rabbit_quorum_queue is enabled, '
+               'Default 0 which means dont set a limit.'),
     cfg.IntOpt('rabbit_transient_queues_ttl',
-               min=1,
+               min=0,
                default=1800,
                help='Positive integer representing duration in seconds for '
                     'queue TTL (x-expires). Queues which are unused for the '
                     'duration of the TTL are automatically deleted. The '
-                    'parameter affects only reply and fanout queues.'),
+                    'parameter affects only reply and fanout queues. Setting '
+                    '0 as value will disable the x-expires. If doing so, '
+                    'make sure you have a rabbitmq policy to delete the '
+                    'queues or you deployment will create an infinite number '
+                    'of queue over time.'
+                    'In case rabbit_stream_fanout is set to True, this option '
+                    'will control data retention policy (x-max-age) for '
+                    'messages in the fanout queue rather then the queue '
+                    'duration itself. So the oldest data in the stream queue '
+                    'will be discarded from it once reaching TTL '
+                    'Setting to 0 will disable x-max-age for stream which '
+                    'make stream grow indefinitely filling up the diskspace'),
     cfg.IntOpt('rabbit_qos_prefetch_count',
                default=0,
                help='Specifies the number of messages to prefetch. Setting to '
@@ -173,7 +236,7 @@ rabbit_opts = [
                "considered down if heartbeat's keep-alive fails "
                "(0 disables heartbeat)."),
     cfg.IntOpt('heartbeat_rate',
-               default=2,
+               default=3,
                help='How often times during the heartbeat_timeout_threshold '
                'we check the heartbeat.'),
     cfg.BoolOpt('direct_mandatory_flag',
@@ -192,14 +255,36 @@ rabbit_opts = [
                 default=False,
                 help="Enable x-cancel-on-ha-failover flag so that "
                      "rabbitmq server will cancel and notify consumers"
-                     "when queue is down")
+                     "when queue is down"),
+    cfg.BoolOpt('use_queue_manager',
+                default=False,
+                help='Should we use consistant queue names or random ones'),
+    cfg.StrOpt('hostname',
+               sample_default='node1.example.com',
+               default=socket.gethostname(),
+               help='Hostname used by queue manager. Defaults to the value '
+               'returned by socket.gethostname().'),
+    cfg.StrOpt('processname',
+               sample_default='nova-api',
+               default=os.path.basename(sys.argv[0]),
+               help='Process name used by queue manager'),
+    cfg.BoolOpt('rabbit_stream_fanout',
+                default=False,
+                help='Use stream queues in RabbitMQ (x-queue-type: stream). '
+                'Streams are a new persistent and replicated data structure '
+                '("queue type") in RabbitMQ which models an append-only log '
+                'with non-destructive consumer semantics. It is available '
+                'as of RabbitMQ 3.9.0. If set this option will replace all '
+                'fanout queues with only one stream queue.'),
 ]
 
 LOG = logging.getLogger(__name__)
 
 
 def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
-                         rabbit_quorum_queue):
+                         rabbit_quorum_queue,
+                         rabbit_quorum_queue_config,
+                         rabbit_stream_fanout):
     """Construct the arguments for declaring a queue.
 
     If the rabbit_ha_queues option is set, we try to declare a mirrored queue
@@ -233,6 +318,18 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
     RabbitMQ 3.8.0. If set this option will conflict with
     the HA queues (``rabbit_ha_queues``) aka mirrored queues,
     in other words HA queues should be disabled.
+
+    rabbit_quorum_queue_config:
+    Quorum queues provides three options to handle message poisoning
+    and limit the resources the quorum queue can use
+    x-delivery-limit number of times the queue will try to deliver
+    a message before it decide to discard it
+    x-max-in-memory-length, x-max-in-memory-bytes control the size
+    of memory used by quorum queue
+
+    If the rabbit_stream_fanout option is set, fanout queues are going to use
+    stream instead of quorum queues. See here:
+      https://www.rabbitmq.com/streams.html
     """
     args = {}
 
@@ -246,16 +343,31 @@ def _get_queue_arguments(rabbit_ha_queues, rabbit_queue_ttl,
 
     if rabbit_quorum_queue:
         args['x-queue-type'] = 'quorum'
+        if rabbit_quorum_queue_config.delivery_limit:
+            args['x-delivery-limit'] = \
+                rabbit_quorum_queue_config.delivery_limit
+        if rabbit_quorum_queue_config.max_memory_length:
+            args['x-max-in-memory-length'] = \
+                rabbit_quorum_queue_config.max_memory_length
+        if rabbit_quorum_queue_config.max_memory_bytes:
+            args['x-max-in-memory-bytes'] = \
+                rabbit_quorum_queue_config.max_memory_bytes
 
     if rabbit_queue_ttl > 0:
         args['x-expires'] = rabbit_queue_ttl * 1000
 
+    if rabbit_stream_fanout:
+        args = {'x-queue-type': 'stream'}
+        if rabbit_queue_ttl > 0:
+            # max-age is a string
+            args['x-max-age'] = f"{rabbit_queue_ttl}s"
+
     return args
 
 
 class RabbitMessage(dict):
     def __init__(self, raw_message):
-        super(RabbitMessage, self).__init__(
+        super().__init__(
             rpc_common.deserialize_msg(raw_message.payload))
         LOG.trace('RabbitMessage.Init: message %s', self)
         self._raw_message = raw_message
@@ -269,13 +381,15 @@ class RabbitMessage(dict):
         self._raw_message.requeue()
 
 
-class Consumer(object):
+class Consumer:
     """Consumer class."""
 
     def __init__(self, exchange_name, queue_name, routing_key, type, durable,
                  exchange_auto_delete, queue_auto_delete, callback,
                  nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0,
-                 enable_cancel_on_failover=False, rabbit_quorum_queue=False):
+                 enable_cancel_on_failover=False, rabbit_quorum_queue=False,
+                 rabbit_quorum_queue_config=QuorumMemConfig(0, 0, 0),
+                 rabbit_stream_fanout=False):
         """Init the Consumer class with the exchange_name, routing_key,
         type, durable auto_delete
         """
@@ -288,17 +402,60 @@ class Consumer(object):
         self.callback = callback
         self.type = type
         self.nowait = nowait
-        self.queue_arguments = _get_queue_arguments(rabbit_ha_queues,
-                                                    rabbit_queue_ttl,
-                                                    rabbit_quorum_queue)
+        rabbit_quorum_queue_config = rabbit_quorum_queue_config or {}
+        self.queue_arguments = _get_queue_arguments(
+            rabbit_ha_queues, rabbit_queue_ttl, rabbit_quorum_queue,
+            rabbit_quorum_queue_config, rabbit_stream_fanout)
         self.queue = None
         self._declared_on = None
         self.exchange = kombu.entity.Exchange(
             name=exchange_name,
-            type=type,
+            type=self.type,
             durable=self.durable,
             auto_delete=self.exchange_auto_delete)
         self.enable_cancel_on_failover = enable_cancel_on_failover
+        self.rabbit_stream_fanout = rabbit_stream_fanout
+        self.next_stream_offset = "last"
+
+    def _declare_fallback(self, err, conn, consumer_arguments):
+        """Fallback by declaring a non durable queue.
+
+        When a control exchange is shared between services it is possible
+        that some service created first a non durable control exchange and
+        then after that an other service can try to create the same control
+        exchange but as a durable control exchange. In this case RabbitMQ
+        will raise an exception (PreconditionFailed), and then it will stop
+        our execution and our service will fail entirly. In this case we want
+        to fallback by creating a non durable queue to match the default
+        config.
+        """
+        if "PRECONDITION_FAILED - inequivalent arg 'durable'" in str(err):
+            LOG.info(
+                "[%s] Retrying to declare the exchange (%s) as "
+                "non durable", conn.connection_id, self.exchange_name)
+            self.exchange = kombu.entity.Exchange(
+                name=self.exchange_name,
+                type=self.type,
+                durable=False,
+                auto_delete=self.queue_auto_delete)
+            self.queue = kombu.entity.Queue(
+                name=self.queue_name,
+                channel=conn.channel,
+                exchange=self.exchange,
+                durable=False,
+                auto_delete=self.queue_auto_delete,
+                routing_key=self.routing_key,
+                queue_arguments=self.queue_arguments,
+                consumer_arguments=consumer_arguments
+            )
+            self.queue.declare()
+
+    def reset_stream_offset(self):
+        if not self.rabbit_stream_fanout:
+            return
+        LOG.warn("Reset consumer for queue %s next offset was at %s.",
+                 self.queue_name, self.next_stream_offset)
+        self.next_stream_offset = "last"
 
     def declare(self, conn):
         """Re-declare the queue after a rabbit (re)connect."""
@@ -308,6 +465,10 @@ class Consumer(object):
             consumer_arguments = {
                 "x-cancel-on-ha-failover": True}
 
+        if self.rabbit_stream_fanout:
+            consumer_arguments = {
+                "x-stream-offset": self.next_stream_offset}
+
         self.queue = kombu.entity.Queue(
             name=self.queue_name,
             channel=conn.channel,
@@ -320,9 +481,50 @@ class Consumer(object):
         )
 
         try:
-            LOG.debug('[%s] Queue.declare: %s',
-                      conn.connection_id, self.queue_name)
-            self.queue.declare()
+            if self.rabbit_stream_fanout:
+                LOG.info('[%s] Stream Queue.declare: %s after offset %s',
+                         conn.connection_id, self.queue_name,
+                         self.next_stream_offset)
+            else:
+                LOG.debug('[%s] Queue.declare: %s',
+                          conn.connection_id, self.queue_name)
+            try:
+                self.queue.declare()
+            except amqp_ex.PreconditionFailed as err:
+                # NOTE(hberaud): This kind of exception may be triggered
+                # when a control exchange is shared between services and
+                # when services try to create it with configs that differ
+                # from each others. RabbitMQ will reject the services
+                # that try to create it with a configuration that differ
+                # from the one used first.
+                LOG.warning('[%s] Queue %s could not be declared probably '
+                            'because of conflicting configurations: %s',
+                            conn.connection_id, self.queue_name, err)
+                self._declare_fallback(err, conn, consumer_arguments)
+            except amqp_ex.NotFound as ex:
+                # NOTE(viktor.krivak): This exception is raised when
+                # non-durable and non-ha queue is hosted on node that
+                # is currently unresponsive or down. Thus, it is
+                # not possible to redeclare the queue since its status
+                # is unknown, and it could disappear in
+                # a few moments. However, the queue could be explicitly
+                # deleted and the redeclaration will then succeed.
+                # Queue will be then scheduled on any of the
+                # running/responsive nodes.
+                # This fixes bug:
+                # https://bugs.launchpad.net/oslo.messaging/+bug/2068630
+
+                LOG.warning("Queue %s is stuck on unresponsive node. "
+                            "Trying to delete the queue and redeclare it "
+                            "again, Error info: %s", self.queue_name, ex)
+                try:
+                    self.queue.delete()
+                except Exception as in_ex:
+                    LOG.warning("During cleanup of stuck queue deletion "
+                                "another exception occurred: %s. Ignoring...",
+                                in_ex)
+                self.queue.declare()
+
         except conn.connection.channel_errors as exc:
             # NOTE(jrosenboom): This exception may be triggered by a race
             # condition. Simply retrying will solve the error most of the time
@@ -351,6 +553,14 @@ class Consumer(object):
                           'Queue: [%(queue)s], '
                           'error message: [%(err_str)s]', info)
                 time.sleep(interval)
+                if self.queue_arguments.get('x-queue-type') == 'quorum':
+                    # Before re-declare queue, try to delete it
+                    # This is helping with issue #2028384
+                    # NOTE(amorin) we need to make sure the connection is
+                    # established again, because when an error occur, the
+                    # connection is closed.
+                    conn.ensure_connection()
+                    self.queue.delete()
                 self.queue.declare()
             else:
                 raise
@@ -393,6 +603,24 @@ class Consumer(object):
                                    nowait=self.nowait)
             else:
                 raise
+        except amqp_ex.InternalError as exc:
+            if self.queue_arguments.get('x-queue-type') == 'quorum':
+                # Before re-consume queue, try to delete it
+                # This is helping with issue #2028384
+                if exc.code == 541:
+                    LOG.warning('Queue %s seems broken, will try delete it '
+                                'before starting over.', self.queue.name)
+                    # NOTE(amorin) we need to make sure the connection is
+                    # established again, because when an error occur, the
+                    # connection is closed.
+                    conn.ensure_connection()
+                    self.queue.delete()
+                    self.declare(conn)
+                    self.queue.consume(callback=self._callback,
+                                       consumer_tag=str(tag),
+                                       nowait=self.nowait)
+            else:
+                raise
 
     def cancel(self, tag):
         LOG.trace('ConsumerBase.cancel: canceling %s', tag)
@@ -403,6 +631,12 @@ class Consumer(object):
 
         Messages that are processed and ack'ed.
         """
+        if self.rabbit_stream_fanout:
+            offset = message.headers.get("x-stream-offset")
+            if offset is not None:
+                LOG.debug("Stream for %s current offset: %s",
+                          self.queue_name, offset)
+                self.next_stream_offset = offset + 1
 
         m2p = getattr(self.queue.channel, 'message_to_python', None)
         if m2p:
@@ -490,7 +724,7 @@ class ConnectionLock(DummyConnectionLock):
             self.release()
 
 
-class Connection(object):
+class Connection:
     """Connection object."""
 
     def __init__(self, conf, url, purpose, retry=None):
@@ -505,6 +739,11 @@ class Connection(object):
         self.login_method = driver_conf.rabbit_login_method
         self.rabbit_ha_queues = driver_conf.rabbit_ha_queues
         self.rabbit_quorum_queue = driver_conf.rabbit_quorum_queue
+        self.rabbit_quorum_queue_config = self._get_quorum_configurations(
+            driver_conf)
+        self.rabbit_transient_quorum_queue = \
+            driver_conf.rabbit_transient_quorum_queue
+        self.rabbit_stream_fanout = driver_conf.rabbit_stream_fanout
         self.rabbit_transient_queues_ttl = \
             driver_conf.rabbit_transient_queues_ttl
         self.rabbit_qos_prefetch_count = driver_conf.rabbit_qos_prefetch_count
@@ -522,6 +761,18 @@ class Connection(object):
         self.heartbeat_in_pthread = driver_conf.heartbeat_in_pthread
         self.ssl_enforce_fips_mode = driver_conf.ssl_enforce_fips_mode
         self.enable_cancel_on_failover = driver_conf.enable_cancel_on_failover
+        self.use_queue_manager = driver_conf.use_queue_manager
+
+        if self.rabbit_stream_fanout and self.rabbit_qos_prefetch_count <= 0:
+            raise RuntimeError('Configuration Error: rabbit_stream_fanout '
+                               'need rabbit_qos_prefetch_count to be set to '
+                               'a value greater than 0.')
+
+        if (self.rabbit_stream_fanout and not
+                self.rabbit_transient_quorum_queue):
+            raise RuntimeError('Configuration Error: rabbit_stream_fanout '
+                               'need rabbit_transient_quorum_queue to be set '
+                               'to true.')
 
         if self.heartbeat_in_pthread:
             # NOTE(hberaud): Experimental: threading module is in use to run
@@ -549,6 +800,8 @@ class Connection(object):
             # if it was already monkey patched by eventlet/greenlet.
             global threading
             threading = _utils.stdlib_threading
+            amqpdriver.threading = _utils.stdlib_threading
+            amqpdriver.queue = _utils.stdlib_queue
 
         self.direct_mandatory_flag = driver_conf.direct_mandatory_flag
 
@@ -699,6 +952,13 @@ class Connection(object):
             self.connection.port = 1234
             self._poll_timeout = 0.05
 
+        if self.use_queue_manager:
+            self._q_manager = amqpdriver.QManager(
+                hostname=driver_conf.hostname,
+                processname=driver_conf.processname)
+        else:
+            self._q_manager = None
+
     # FIXME(markmc): use oslo sslutils when it is available as a library
     _SSL_PROTOCOLS = {
         "tlsv1": ssl.PROTOCOL_TLSv1,
@@ -732,26 +992,28 @@ class Connection(object):
         except KeyError:
             raise RuntimeError("Invalid SSL version : %s" % version)
 
+    def _get_quorum_configurations(self, driver_conf):
+        """Get the quorum queue configurations"""
+        delivery_limit = driver_conf.rabbit_quorum_delivery_limit
+        max_memory_length = driver_conf.rabbit_quorum_max_memory_length
+        max_memory_bytes = driver_conf.rabbit_quorum_max_memory_bytes
+        return QuorumMemConfig(delivery_limit, max_memory_length,
+                               max_memory_bytes)
+
     # NOTE(moguimar): default_password in this function's context is just
     #                 a fallback option, not a hardcoded password.
     def _transform_transport_url(self, url, host, default_username='',  # nosec
                                  default_password='', default_hostname=''):
         transport = url.transport.replace('kombu+', '')
         transport = transport.replace('rabbit', 'amqp')
-        return '%s://%s:%s@%s:%s/%s' % (
+        return '{}://{}:{}@{}:{}/{}'.format(
             transport,
             parse.quote(host.username or default_username),
             parse.quote(host.password or default_password),
-            self._parse_url_hostname(host.hostname) or default_hostname,
+            netutils.escape_ipv6(host.hostname) or default_hostname,
             str(host.port or 5672),
             url.virtual_host or '')
 
-    def _parse_url_hostname(self, hostname):
-        """Handles hostname returned from urlparse and checks whether it's
-        ipaddress. If it's ipaddress it ensures that it has brackets for IPv6.
-        """
-        return '[%s]' % hostname if ':' in hostname else hostname
-
     def _fetch_ssl_params(self):
         """Handles fetching what ssl params should be used for the connection
         (if any).
@@ -759,7 +1021,7 @@ class Connection(object):
         if self.ssl:
             ssl_params = dict()
 
-            # http://docs.python.org/library/ssl.html - ssl.wrap_socket
+            # http://docs.python.org/library/ssl.html
             if self.ssl_version:
                 ssl_params['ssl_version'] = self.validate_ssl_version(
                     self.ssl_version)
@@ -837,11 +1099,24 @@ class Connection(object):
             info = {'err_str': exc, 'sleep_time': interval}
             info.update(self._get_connection_info(conn_error=True))
 
+            if self.rabbit_stream_fanout and 'Basic.cancel' in str(exc):
+                # This branch allows for consumer offset reset
+                # in the unlikely case consumers are cancelled. This may
+                # happen, for example, when we delete the stream queue.
+                # We need to start consuming from "last" because the stream
+                # offset maybe reset.
+                LOG.warn('[%s] Basic.cancel received. '
+                         'Resetting consumers offsets to last.',
+                         self.connection_id)
+                for consumer in self._consumers:
+                    consumer.reset_stream_offset()
+
             if 'Socket closed' in str(exc):
                 LOG.error('[%(connection_id)s] AMQP server'
                           ' %(hostname)s:%(port)s closed'
                           ' the connection. Check login credentials:'
                           ' %(err_str)s', info)
+
             else:
                 LOG.error('[%(connection_id)s] AMQP server on '
                           '%(hostname)s:%(port)s is unreachable: '
@@ -951,11 +1226,20 @@ class Connection(object):
         """Close/release this connection."""
         self._heartbeat_stop()
         if self.connection:
-            for consumer in filter(lambda c: c.type == 'fanout',
-                                   self._consumers):
-                LOG.debug('[connection close] Deleting fanout '
-                          'queue: %s ' % consumer.queue.name)
-                consumer.queue.delete()
+            # NOTE(jcosmao) Delete queue should be called only when queue name
+            # is randomized. When using streams, queue is shared between
+            # all consumers, thus deleting fanout queue will force all other
+            # consumers to disconnect/reconnect by throwing
+            # amqp.exceptions.ConsumerCancelled.
+            # When using QManager, queue name is consistent accross agent
+            # restart, so we don't need to delete it either. Deletion must be
+            # handled by expiration policy.
+            if not self.rabbit_stream_fanout and not self.use_queue_manager:
+                for consumer in filter(lambda c: c.type == 'fanout',
+                                       self._consumers):
+                    LOG.debug('[connection close] Deleting fanout '
+                              'queue: %s ' % consumer.queue.name)
+                    consumer.queue.delete()
             self._set_current_channel(None)
             self.connection.release()
             self.connection = None
@@ -1025,7 +1309,7 @@ class Connection(object):
                     sock.setsockopt(socket.IPPROTO_TCP,
                                     TCP_USER_TIMEOUT,
                                     int(math.ceil(timeout)))
-                except socket.error as error:
+                except OSError as error:
                     code = error[0]
                     # TCP_USER_TIMEOUT not defined on kernels <2.6.37
                     if code != errno.ENOPROTOOPT:
@@ -1090,7 +1374,7 @@ class Connection(object):
                             ConnectionRefusedError,
                             OSError,
                             kombu.exceptions.OperationalError,
-                            amqp_exec.ConnectionForced) as exc:
+                            amqp_ex.ConnectionForced) as exc:
                         LOG.info("A recoverable connection/channel error "
                                  "occurred, trying to reconnect: %s", exc)
                         self.ensure_connection()
@@ -1201,13 +1485,15 @@ class Connection(object):
             queue_name=topic,
             routing_key='',
             type='direct',
-            durable=False,
+            durable=self.rabbit_transient_quorum_queue,
             exchange_auto_delete=False,
             queue_auto_delete=False,
             callback=callback,
             rabbit_ha_queues=self.rabbit_ha_queues,
             rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
-            enable_cancel_on_failover=self.enable_cancel_on_failover)
+            enable_cancel_on_failover=self.enable_cancel_on_failover,
+            rabbit_quorum_queue=self.rabbit_transient_quorum_queue,
+            rabbit_quorum_queue_config=self.rabbit_quorum_queue_config)
 
         self.declare_consumer(consumer)
 
@@ -1225,29 +1511,43 @@ class Connection(object):
             callback=callback,
             rabbit_ha_queues=self.rabbit_ha_queues,
             enable_cancel_on_failover=self.enable_cancel_on_failover,
-            rabbit_quorum_queue=self.rabbit_quorum_queue)
+            rabbit_quorum_queue=self.rabbit_quorum_queue,
+            rabbit_quorum_queue_config=self.rabbit_quorum_queue_config)
 
         self.declare_consumer(consumer)
 
     def declare_fanout_consumer(self, topic, callback):
         """Create a 'fanout' consumer."""
 
-        unique = uuid.uuid4().hex
         exchange_name = '%s_fanout' % topic
-        queue_name = '%s_fanout_%s' % (topic, unique)
+        if self.rabbit_stream_fanout:
+            queue_name = '%s_fanout' % topic
+        else:
+            if self._q_manager:
+                unique = self._q_manager.get()
+            else:
+                unique = uuid.uuid4().hex
+            queue_name = '{}_fanout_{}'.format(topic, unique)
+        LOG.debug('Creating fanout queue: %s', queue_name)
+
+        is_durable = (self.rabbit_transient_quorum_queue or
+                      self.rabbit_stream_fanout)
 
         consumer = Consumer(
             exchange_name=exchange_name,
             queue_name=queue_name,
             routing_key=topic,
             type='fanout',
-            durable=False,
+            durable=is_durable,
             exchange_auto_delete=True,
             queue_auto_delete=False,
             callback=callback,
             rabbit_ha_queues=self.rabbit_ha_queues,
             rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
-            enable_cancel_on_failover=self.enable_cancel_on_failover)
+            enable_cancel_on_failover=self.enable_cancel_on_failover,
+            rabbit_quorum_queue=self.rabbit_transient_quorum_queue,
+            rabbit_quorum_queue_config=self.rabbit_quorum_queue_config,
+            rabbit_stream_fanout=self.rabbit_stream_fanout)
 
         self.declare_consumer(consumer)
 
@@ -1273,8 +1573,8 @@ class Connection(object):
         # the connection's socket while it is in an error state will cause
         # py-amqp to attempt reconnecting.
         ci = self.connection.info()
-        info = dict([(k, ci.get(k)) for k in
-                     ['hostname', 'port', 'transport']])
+        info = {k: ci.get(k) for k in
+                ['hostname', 'port', 'transport']}
         client_port = None
         if (not conn_error and self.channel and
                 hasattr(self.channel.connection, 'sock') and
@@ -1289,7 +1589,20 @@ class Connection(object):
         """Publish a message."""
 
         if not (exchange.passive or exchange.name in self._declared_exchanges):
-            exchange(self.channel).declare()
+            try:
+                exchange(self.channel).declare()
+            except amqp_ex.PreconditionFailed as err:
+                # NOTE(hberaud): This kind of exception may be triggered
+                # when a control exchange is shared between services and
+                # when services try to create it with configs that differ
+                # from each others. RabbitMQ will reject the services
+                # that try to create it with a configuration that differ
+                # from the one used first.
+                if "PRECONDITION_FAILED - inequivalent arg 'durable'" \
+                   in str(err):
+                    LOG.warning("Force creating a non durable exchange.")
+                    exchange.durable = False
+                    exchange(self.channel).declare()
             self._declared_exchanges.add(exchange.name)
 
         log_info = {'msg': msg,
@@ -1340,7 +1653,9 @@ class Connection(object):
                 queue_arguments=_get_queue_arguments(
                     self.rabbit_ha_queues,
                     0,
-                    self.rabbit_quorum_queue))
+                    self.rabbit_quorum_queue,
+                    self.rabbit_quorum_queue_config,
+                    False))
             log_info = {'key': routing_key, 'exchange': exchange}
             LOG.trace(
                 'Connection._publish_and_creates_default_queue: '
@@ -1379,13 +1694,16 @@ class Connection(object):
 
     def direct_send(self, msg_id, msg):
         """Send a 'direct' message."""
-        exchange = kombu.entity.Exchange(name='',  # using default exchange
-                                         type='direct',
-                                         durable=False,
-                                         auto_delete=True,
-                                         passive=True)
+        exchange = kombu.entity.Exchange(
+            name='',  # using default exchange
+            type='direct',
+            durable=self.rabbit_transient_quorum_queue,
+            auto_delete=True,
+            passive=True)
         options = oslo_messaging.TransportOptions(
             at_least_once=self.direct_mandatory_flag)
+
+        LOG.debug('Sending direct to %s', msg_id)
         self._ensure_publishing(self._publish_and_raises_on_missing_exchange,
                                 exchange, msg, routing_key=msg_id,
                                 transport_options=options)
@@ -1399,6 +1717,8 @@ class Connection(object):
             durable=self.durable,
             auto_delete=self.amqp_auto_delete)
 
+        LOG.debug('Sending topic to %s with routing_key %s', exchange_name,
+                  topic)
         self._ensure_publishing(self._publish, exchange, msg,
                                 routing_key=topic, timeout=timeout,
                                 retry=retry,
@@ -1406,11 +1726,13 @@ class Connection(object):
 
     def fanout_send(self, topic, msg, retry=None):
         """Send a 'fanout' message."""
-        exchange = kombu.entity.Exchange(name='%s_fanout' % topic,
-                                         type='fanout',
-                                         durable=False,
-                                         auto_delete=True)
+        exchange = kombu.entity.Exchange(
+            name='%s_fanout' % topic,
+            type='fanout',
+            durable=self.rabbit_transient_quorum_queue,
+            auto_delete=True)
 
+        LOG.debug('Sending fanout to %s_fanout', topic)
         self._ensure_publishing(self._publish, exchange, msg, retry=retry)
 
     def notify_send(self, exchange_name, topic, msg, retry=None, **kwargs):
@@ -1466,7 +1788,7 @@ class RabbitDriver(amqpdriver.AMQPDriverBase):
             conf, max_size, min_size, ttl,
             url, Connection)
 
-        super(RabbitDriver, self).__init__(
+        super().__init__(
             conf, url,
             connection_pool,
             default_exchange,
