@@ -1,4 +1,3 @@
-
 # Copyright 2011 OpenStack Foundation.
 # All Rights Reserved.
 # Copyright 2013 Red Hat, Inc.
@@ -22,11 +21,9 @@ import uuid
 
 from oslo_config import cfg
 from oslo_utils import timeutils
-import six
 from stevedore import extension
 from stevedore import named
 
-from oslo_messaging._i18n import _LE
 from oslo_messaging import serializer as msg_serializer
 from oslo_messaging import transport as msg_transport
 
@@ -95,7 +92,8 @@ def _send_notification():
                             'oslo.messaging.notify.drivers'
                         ).names(),
                         help='Notification driver')
-    parser.add_argument('payload')
+    parser.add_argument('payload',
+                        help="the notification payload (dict)")
     args = parser.parse_args()
     conf = cfg.ConfigOpts()
     conf([],
@@ -106,8 +104,7 @@ def _send_notification():
     notifier._notify({}, args.event_type, args.payload, args.priority)
 
 
-@six.add_metaclass(abc.ABCMeta)
-class Driver(object):
+class Driver(metaclass=abc.ABCMeta):
     """Base driver for Notifications"""
 
     def __init__(self, conf, topics, transport):
@@ -140,8 +137,7 @@ class Driver(object):
         pass
 
 
-def get_notification_transport(conf, url=None,
-                               allowed_remote_exmods=None, aliases=None):
+def get_notification_transport(conf, url=None, allowed_remote_exmods=None):
     """A factory method for Transport objects for notifications.
 
     This method should be used for notifications, in case notifications are
@@ -158,24 +154,41 @@ def get_notification_transport(conf, url=None,
 
     :param conf: the user configuration
     :type conf: cfg.ConfigOpts
-    :param url: a transport URL
+    :param url: a transport URL, see :py:class:`transport.TransportURL`
     :type url: str or TransportURL
     :param allowed_remote_exmods: a list of modules which a client using this
                                   transport will deserialize remote exceptions
                                   from
     :type allowed_remote_exmods: list
-    :param aliases: A map of transport alias to transport name
-    :type aliases: dict
     """
     conf.register_opts(_notifier_opts,
                        group='oslo_messaging_notifications')
     if url is None:
         url = conf.oslo_messaging_notifications.transport_url
-    return msg_transport.get_transport(conf, url,
-                                       allowed_remote_exmods, aliases)
+    return msg_transport._get_transport(
+        conf, url, allowed_remote_exmods,
+        transport_cls=msg_transport.NotificationTransport)
 
 
-class Notifier(object):
+def _sanitize_context(ctxt):
+    if ctxt is None or type(ctxt) is dict:
+        # NOTE(JayF): Logging drivers, unit tests, and some code calls
+        #             notifier with an emptydict or None instead of an
+        #             actual context. In these cases, discard the passed
+        #             value.
+        return {}
+
+    try:
+        return ctxt.redacted_copy()
+    except AttributeError:
+        # NOTE(JayF): We'd rather send a notification without any context
+        #             than missing sending the notification altogether.
+        _LOG.warning("Unable to properly redact context for "
+                     "notification, omitting context from notification.")
+        return {}
+
+
+class Notifier:
 
     """Send notification messages.
 
@@ -184,7 +197,7 @@ class Notifier(object):
 
     Notification messages follow the following format::
 
-        {'message_id': six.text_type(uuid.uuid4()),
+        {'message_id': str(uuid.uuid4()),
          'publisher_id': 'compute.host1',
          'timestamp': timeutils.utcnow(),
          'priority': 'WARN',
@@ -192,7 +205,7 @@ class Notifier(object):
          'payload': {'instance_id': 12, ... }}
 
     A Notifier object can be instantiated with a transport object and a
-    publisher ID:
+    publisher ID::
 
         notifier = messaging.Notifier(get_notification_transport(CONF),
                                       'compute')
@@ -245,6 +258,10 @@ class Notifier(object):
         conf.register_opts(_notifier_opts,
                            group='oslo_messaging_notifications')
 
+        if not isinstance(transport, msg_transport.NotificationTransport):
+            _LOG.warning("Using RPC transport for notifications. Please use "
+                         "get_notification_transport to obtain a "
+                         "notification transport instance.")
         self.transport = transport
         self.publisher_id = publisher_id
         if retry is not None:
@@ -296,22 +313,27 @@ class Notifier(object):
     def _notify(self, ctxt, event_type, payload, priority, publisher_id=None,
                 retry=None):
         payload = self._serializer.serialize_entity(ctxt, payload)
-        ctxt = self._serializer.serialize_context(ctxt)
 
-        msg = dict(message_id=six.text_type(uuid.uuid4()),
+        # NOTE(JayF): We must remove secure information from notification
+        #             payloads, otherwise we risk sending sensitive creds
+        #             to a notification bus.
+        safe_ctxt = _sanitize_context(ctxt)
+        ctxt = self._serializer.serialize_context(safe_ctxt)
+
+        msg = dict(message_id=str(uuid.uuid4()),
                    publisher_id=publisher_id or self.publisher_id,
                    event_type=event_type,
                    priority=priority,
                    payload=payload,
-                   timestamp=six.text_type(timeutils.utcnow()))
+                   timestamp=str(timeutils.utcnow()))
 
         def do_notify(ext):
             try:
                 ext.obj.notify(ctxt, msg, priority, retry or self.retry)
             except Exception as e:
-                _LOG.exception(_LE("Problem '%(e)s' attempting to send to "
-                                   "notification system. Payload=%(payload)s"),
-                               dict(e=e, payload=payload))
+                _LOG.exception("Problem '%(e)s' attempting to send to "
+                               "notification system. Payload=%(payload)s",
+                               {'e': e, 'payload': payload})
 
         if self._driver_mgr.extensions:
             self._driver_mgr.map(do_notify)
@@ -445,7 +467,7 @@ class _SubNotifier(Notifier):
         self._driver_mgr = self._base._driver_mgr
 
     def _notify(self, ctxt, event_type, payload, priority):
-        super(_SubNotifier, self)._notify(ctxt, event_type, payload, priority)
+        super()._notify(ctxt, event_type, payload, priority)
 
     @classmethod
     def _prepare(cls, base, publisher_id=_marker, retry=_marker):
