@@ -16,13 +16,6 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-__all__ = [
-    'ExecutorLoadFailure',
-    'MessageHandlingServer',
-    'MessagingServerError',
-    'ServerListenError',
-]
-
 import abc
 import functools
 import inspect
@@ -34,13 +27,19 @@ from oslo_config import cfg
 from oslo_service import service
 from oslo_utils import eventletutils
 from oslo_utils import timeutils
-import six
 from stevedore import driver
 
 from oslo_messaging._drivers import base as driver_base
-from oslo_messaging._i18n import _LW, _LI
+from oslo_messaging import _utils as utils
 from oslo_messaging import exceptions
 
+__all__ = [
+    'ExecutorLoadFailure',
+    'MessageHandlingServer',
+    'MessagingServerError',
+    'ServerListenError',
+]
+
 LOG = logging.getLogger(__name__)
 
 # The default number of seconds of waiting after which we will emit a log
@@ -65,8 +64,8 @@ class ExecutorLoadFailure(MessagingServerError):
     """Raised if an executor can't be loaded."""
 
     def __init__(self, executor, ex):
-        msg = 'Failed to load executor "%s": %s' % (executor, ex)
-        super(ExecutorLoadFailure, self).__init__(msg)
+        msg = 'Failed to load executor "{}": {}'.format(executor, ex)
+        super().__init__(msg)
         self.executor = executor
         self.ex = ex
 
@@ -75,8 +74,8 @@ class ServerListenError(MessagingServerError):
     """Raised if we failed to listen on a target."""
 
     def __init__(self, target, ex):
-        msg = 'Failed to listen on target "%s": %s' % (target, ex)
-        super(ServerListenError, self).__init__(msg)
+        msg = 'Failed to listen on target "{}": {}'.format(target, ex)
+        super().__init__(msg)
         self.target = target
         self.ex = ex
 
@@ -85,7 +84,7 @@ class TaskTimeout(MessagingServerError):
     """Raised if we timed out waiting for a task to complete."""
 
 
-class _OrderedTask(object):
+class _OrderedTask:
     """A task which must be executed in a particular order.
 
     A caller may wait for this task to complete by calling
@@ -106,7 +105,7 @@ class _OrderedTask(object):
 
         :param name: The name of this task. Used in log messages.
         """
-        super(_OrderedTask, self).__init__()
+        super().__init__()
 
         self._name = name
         self._cond = threading.Condition()
@@ -125,7 +124,7 @@ class _OrderedTask(object):
 
         while condition():
             if log_timer is not None and log_timer.expired():
-                LOG.warning(_LW('Possible hang: %s'), msg)
+                LOG.warning('Possible hang: %s', msg)
                 LOG.debug(''.join(traceback.format_stack()))
                 # Only log once. After than we wait indefinitely without
                 # logging.
@@ -159,7 +158,7 @@ class _OrderedTask(object):
                               `timeout_timer` expires while waiting.
         """
         with self._cond:
-            msg = '%s is waiting for %s to complete' % (caller, self._name)
+            msg = '{} is waiting for {} to complete'.format(caller, self._name)
             self._wait(lambda: not self.complete,
                        msg, log_after, timeout_timer)
 
@@ -207,11 +206,11 @@ class _OrderedTask(object):
                            msg, log_after, timeout_timer)
 
 
-class _OrderedTaskRunner(object):
+class _OrderedTaskRunner:
     """Mixin for a class which executes ordered tasks."""
 
     def __init__(self, *args, **kwargs):
-        super(_OrderedTaskRunner, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
         # Get a list of methods on this object which have the _ordered
         # attribute
@@ -297,8 +296,8 @@ def ordered(after=None, reset_after=None):
     return _ordered
 
 
-@six.add_metaclass(abc.ABCMeta)
-class MessageHandlingServer(service.ServiceBase, _OrderedTaskRunner):
+class MessageHandlingServer(service.ServiceBase, _OrderedTaskRunner,
+                            metaclass=abc.ABCMeta):
     """Server for handling messages.
 
     Connect a transport to a dispatcher that knows how to process the
@@ -306,16 +305,17 @@ class MessageHandlingServer(service.ServiceBase, _OrderedTaskRunner):
     new tasks.
     """
 
-    def __init__(self, transport, dispatcher, executor='blocking'):
+    def __init__(self, transport, dispatcher, executor=None):
         """Construct a message handling server.
 
         The dispatcher parameter is a DispatcherBase instance which is used
         for routing request to endpoint for processing.
 
         The executor parameter controls how incoming messages will be received
-        and dispatched. By default, the most simple executor is used - the
-        blocking executor. It handles only one message at once. It's
-        recommended to use threading or eventlet.
+        and dispatched. Executor is automatically detected from
+        execution environment.
+        It handles many message in parallel. If your application need
+        asynchronism then you need to consider to use the eventlet executor.
 
         :param transport: the messaging transport
         :type transport: Transport
@@ -323,24 +323,23 @@ class MessageHandlingServer(service.ServiceBase, _OrderedTaskRunner):
                            incoming request
         :type dispatcher: DispatcherBase
         :param executor: name of message executor - available values are
-                         'eventlet', 'blocking' and 'threading'
+                         'eventlet' and 'threading'
         :type executor: str
         """
+        if executor and executor not in ("threading", "eventlet"):
+            raise ExecutorLoadFailure(
+                executor,
+                "Executor should be None or 'eventlet' and 'threading'")
+        if not executor:
+            executor = utils.get_executor_with_context()
+
         self.conf = transport.conf
         self.conf.register_opts(_pool_opts)
 
         self.transport = transport
         self.dispatcher = dispatcher
         self.executor_type = executor
-        if self.executor_type == 'blocking':
-            # NOTE(sileht): We keep blocking as default to not enforce the
-            # application to use threading or eventlet. Because application
-            # have to be preprepared accordingly for each one (monkeypatching,
-            # threadsafe, ...)
-            LOG.info(_LI("blocking executor handles only one message at "
-                         "once. threading or eventlet executor is "
-                         "recommended."))
-        elif self.executor_type == "eventlet":
+        if self.executor_type == "eventlet":
             eventletutils.warn_eventlet_not_patched(
                 expected_patched_modules=['thread'],
                 what="the 'oslo.messaging eventlet executor'")
@@ -359,7 +358,7 @@ class MessageHandlingServer(service.ServiceBase, _OrderedTaskRunner):
 
         self._started = False
 
-        super(MessageHandlingServer, self).__init__()
+        super().__init__()
 
     def _on_incoming(self, incoming):
         """Handles on_incoming event
@@ -396,21 +395,18 @@ class MessageHandlingServer(service.ServiceBase, _OrderedTaskRunner):
         choose to dispatch messages in a new thread, coroutine or simply the
         current thread.
         """
-        # Warn that restarting will be deprecated
         if self._started:
-            LOG.warning(_LW('Restarting a MessageHandlingServer is inherently '
-                            'racy. It is deprecated, and will become a noop '
-                            'in a future release of oslo.messaging. If you '
-                            'need to restart MessageHandlingServer you should '
-                            'instantiate a new object.'))
+            LOG.warning('The server has already been started. Ignoring '
+                        'the redundant call to start().')
+            return
+
         self._started = True
 
         executor_opts = {}
 
-        if self.executor_type in ("threading", "eventlet"):
-            executor_opts["max_workers"] = (
-                override_pool_size or self.conf.executor_thread_pool_size
-            )
+        executor_opts["max_workers"] = (
+            override_pool_size or self.conf.executor_thread_pool_size
+        )
         self._work_executor = self._executor_cls(**executor_opts)
 
         try:
@@ -429,7 +425,8 @@ class MessageHandlingServer(service.ServiceBase, _OrderedTaskRunner):
         some messages, and underlying driver resources associated to this
         server are still in use. See 'wait' for more details.
         """
-        self.listener.stop()
+        if self.listener:
+            self.listener.stop()
         self._started = False
 
     @ordered(after='stop')
@@ -446,7 +443,8 @@ class MessageHandlingServer(service.ServiceBase, _OrderedTaskRunner):
         self._work_executor.shutdown(wait=True)
 
         # Close listener connection after processing all messages
-        self.listener.cleanup()
+        if self.listener:
+            self.listener.cleanup()
 
     def reset(self):
         """Reset service.
