@@ -12,12 +12,13 @@
 #    under the License.
 
 import os
+import requests
+import subprocess
 import time
 import uuid
 
 import concurrent.futures
 from oslo_config import cfg
-import six.moves
 from testtools import matchers
 
 import oslo_messaging
@@ -27,8 +28,8 @@ from oslo_messaging.tests.functional import utils
 class CallTestCase(utils.SkipIfNoTransportURL):
 
     def setUp(self):
-        super(CallTestCase, self).setUp(conf=cfg.ConfigOpts())
-        if self.url.startswith("kafka://"):
+        super().setUp(conf=cfg.ConfigOpts())
+        if self.rpc_url.startswith("kafka://"):
             self.skipTest("kafka does not support RPC API")
 
         self.conf.prog = "test_prog"
@@ -39,7 +40,7 @@ class CallTestCase(utils.SkipIfNoTransportURL):
 
     def test_specific_server(self):
         group = self.useFixture(utils.RpcServerGroupFixture(
-            self.conf, self.url)
+            self.conf, self.rpc_url)
         )
         client = group.client(1)
         client.append(text='open')
@@ -55,7 +56,7 @@ class CallTestCase(utils.SkipIfNoTransportURL):
 
     def test_server_in_group(self):
         group = self.useFixture(
-            utils.RpcServerGroupFixture(self.conf, self.url)
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url)
         )
 
         client = group.client()
@@ -73,13 +74,13 @@ class CallTestCase(utils.SkipIfNoTransportURL):
         # teardown may hang unless we broadcast all control messages
         # to each server
         group1 = self.useFixture(
-            utils.RpcServerGroupFixture(self.conf, self.url,
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url,
                                         use_fanout_ctrl=True))
         group2 = self.useFixture(
-            utils.RpcServerGroupFixture(self.conf, self.url, exchange="a",
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url, exchange="a",
                                         use_fanout_ctrl=True))
         group3 = self.useFixture(
-            utils.RpcServerGroupFixture(self.conf, self.url, exchange="b",
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url, exchange="b",
                                         use_fanout_ctrl=True))
 
         client1 = group1.client(1)
@@ -113,7 +114,7 @@ class CallTestCase(utils.SkipIfNoTransportURL):
 
     def test_timeout(self):
         transport = self.useFixture(
-            utils.RPCTransportFixture(self.conf, self.url)
+            utils.RPCTransportFixture(self.conf, self.rpc_url)
         )
         target = oslo_messaging.Target(topic="no_such_topic")
         c = utils.ClientStub(transport.transport, target, timeout=1)
@@ -122,7 +123,7 @@ class CallTestCase(utils.SkipIfNoTransportURL):
 
     def test_exception(self):
         group = self.useFixture(
-            utils.RpcServerGroupFixture(self.conf, self.url)
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url)
         )
         client = group.client(1)
         client.add(increment=2)
@@ -130,12 +131,12 @@ class CallTestCase(utils.SkipIfNoTransportURL):
 
     def test_timeout_with_concurrently_queues(self):
         transport = self.useFixture(
-            utils.RPCTransportFixture(self.conf, self.url)
+            utils.RPCTransportFixture(self.conf, self.rpc_url)
         )
         target = oslo_messaging.Target(topic="topic_" + str(uuid.uuid4()),
                                        server="server_" + str(uuid.uuid4()))
         server = self.useFixture(
-            utils.RpcServerFixture(self.conf, self.url, target,
+            utils.RpcServerFixture(self.conf, self.rpc_url, target,
                                    executor="threading"))
         client = utils.ClientStub(transport.transport, target,
                                   cast=False, timeout=5)
@@ -153,11 +154,11 @@ class CallTestCase(utils.SkipIfNoTransportURL):
         self.assertEqual(10, server.endpoint.ival)
 
     def test_mandatory_call(self):
-        if not self.url.startswith("rabbit://"):
+        if not self.rpc_url.startswith("rabbit://"):
             self.skipTest("backend does not support call monitoring")
 
         transport = self.useFixture(utils.RPCTransportFixture(self.conf,
-                                                              self.url))
+                                                              self.rpc_url))
         target = oslo_messaging.Target(topic='topic_' + str(uuid.uuid4()),
                                        server='server_' + str(uuid.uuid4()))
 
@@ -187,22 +188,22 @@ class CallTestCase(utils.SkipIfNoTransportURL):
                           client2.delay)
 
     def test_monitor_long_call(self):
-        if not (self.url.startswith("rabbit://") or
-                self.url.startswith("amqp://")):
+        if not (self.rpc_url.startswith("rabbit://") or
+                self.rpc_url.startswith("amqp://")):
             self.skipTest("backend does not support call monitoring")
 
         transport = self.useFixture(utils.RPCTransportFixture(self.conf,
-                                                              self.url))
+                                                              self.rpc_url))
         target = oslo_messaging.Target(topic='topic_' + str(uuid.uuid4()),
                                        server='server_' + str(uuid.uuid4()))
 
-        class _endpoint(object):
+        class _endpoint:
             def delay(self, ctxt, seconds):
                 time.sleep(seconds)
                 return seconds
 
         self.useFixture(
-            utils.RpcServerFixture(self.conf, self.url, target,
+            utils.RpcServerFixture(self.conf, self.rpc_url, target,
                                    executor='threading',
                                    endpoint=_endpoint()))
 
@@ -230,7 +231,7 @@ class CallTestCase(utils.SkipIfNoTransportURL):
                                        namespace="Name1",
                                        version="7.5")
 
-        class _endpoint(object):
+        class _endpoint:
             def __init__(self, target):
                 self.target = target()
 
@@ -238,10 +239,10 @@ class CallTestCase(utils.SkipIfNoTransportURL):
                 return echo
 
         transport = self.useFixture(
-            utils.RPCTransportFixture(self.conf, self.url)
+            utils.RPCTransportFixture(self.conf, self.rpc_url)
         )
         self.useFixture(
-            utils.RpcServerFixture(self.conf, self.url, target,
+            utils.RpcServerFixture(self.conf, self.rpc_url, target,
                                    executor="threading",
                                    endpoint=_endpoint(target)))
         client1 = utils.ClientStub(transport.transport, target,
@@ -273,14 +274,14 @@ class CallTestCase(utils.SkipIfNoTransportURL):
     def test_bad_endpoint(self):
         # 'target' attribute is reserved and should be of type Target
 
-        class _endpoint(object):
+        class _endpoint:
             def target(self, ctxt, echo):
                 return echo
 
         target = oslo_messaging.Target(topic="topic_" + str(uuid.uuid4()),
                                        server="server_" + str(uuid.uuid4()))
         transport = self.useFixture(
-            utils.RPCTransportFixture(self.conf, self.url)
+            utils.RPCTransportFixture(self.conf, self.rpc_url)
         )
         self.assertRaises(TypeError,
                           oslo_messaging.get_rpc_server,
@@ -296,13 +297,13 @@ class CastTestCase(utils.SkipIfNoTransportURL):
     # making the necessary assertions.
 
     def setUp(self):
-        super(CastTestCase, self).setUp()
-        if self.url.startswith("kafka://"):
+        super().setUp()
+        if self.rpc_url.startswith("kafka://"):
             self.skipTest("kafka does not support RPC API")
 
     def test_specific_server(self):
         group = self.useFixture(
-            utils.RpcServerGroupFixture(self.conf, self.url)
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url)
         )
         client = group.client(1, cast=True)
         client.append(text='open')
@@ -321,10 +322,10 @@ class CastTestCase(utils.SkipIfNoTransportURL):
             self.assertEqual(0, group.servers[i].endpoint.ival)
 
     def test_server_in_group(self):
-        if self.url.startswith("amqp:"):
+        if self.rpc_url.startswith("amqp:"):
             self.skipTest("QPID-6307")
         group = self.useFixture(
-            utils.RpcServerGroupFixture(self.conf, self.url)
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url)
         )
         client = group.client(cast=True)
         for i in range(20):
@@ -343,7 +344,7 @@ class CastTestCase(utils.SkipIfNoTransportURL):
 
     def test_fanout(self):
         group = self.useFixture(
-            utils.RpcServerGroupFixture(self.conf, self.url)
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url)
         )
         client = group.client('all', cast=True)
         client.append(text='open')
@@ -367,13 +368,14 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
 
     def test_simple(self):
         get_timeout = 1
-        if self.url.startswith("kafka://"):
+        if self.notify_url.startswith("kafka://"):
             get_timeout = 5
             self.conf.set_override('consumer_group', 'test_simple',
                                    group='oslo_messaging_kafka')
 
         listener = self.useFixture(
-            utils.NotificationFixture(self.conf, self.url, ['test_simple']))
+            utils.NotificationFixture(self.conf, self.notify_url,
+                                      ['test_simple']))
         notifier = listener.notifier('abc')
 
         notifier.info({}, 'test', 'Hello World!')
@@ -385,13 +387,13 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
 
     def test_multiple_topics(self):
         get_timeout = 1
-        if self.url.startswith("kafka://"):
+        if self.notify_url.startswith("kafka://"):
             get_timeout = 5
             self.conf.set_override('consumer_group', 'test_multiple_topics',
                                    group='oslo_messaging_kafka')
 
         listener = self.useFixture(
-            utils.NotificationFixture(self.conf, self.url, ['a', 'b']))
+            utils.NotificationFixture(self.conf, self.notify_url, ['a', 'b']))
         a = listener.notifier('pub-a', topics=['a'])
         b = listener.notifier('pub-b', topics=['b'])
 
@@ -416,9 +418,9 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
 
     def test_multiple_servers(self):
         timeout = 0.5
-        if self.url.startswith("amqp:"):
+        if self.notify_url.startswith("amqp:"):
             self.skipTest("QPID-6307")
-        if self.url.startswith("kafka://"):
+        if self.notify_url.startswith("kafka://"):
             self.skipTest("Kafka: needs to be fixed")
             timeout = 5
             self.conf.set_override('consumer_group',
@@ -426,10 +428,12 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
                                    group='oslo_messaging_kafka')
 
         listener_a = self.useFixture(
-            utils.NotificationFixture(self.conf, self.url, ['test-topic']))
+            utils.NotificationFixture(self.conf, self.notify_url,
+                                      ['test-topic']))
 
         listener_b = self.useFixture(
-            utils.NotificationFixture(self.conf, self.url, ['test-topic']))
+            utils.NotificationFixture(self.conf, self.notify_url,
+                                      ['test-topic']))
 
         n = listener_a.notifier('pub')
 
@@ -446,20 +450,20 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
 
     def test_independent_topics(self):
         get_timeout = 0.5
-        if self.url.startswith("kafka://"):
+        if self.notify_url.startswith("kafka://"):
             get_timeout = 5
             self.conf.set_override('consumer_group',
                                    'test_independent_topics_a',
                                    group='oslo_messaging_kafka')
         listener_a = self.useFixture(
-            utils.NotificationFixture(self.conf, self.url, ['1']))
+            utils.NotificationFixture(self.conf, self.notify_url, ['1']))
 
-        if self.url.startswith("kafka://"):
+        if self.notify_url.startswith("kafka://"):
             self.conf.set_override('consumer_group',
                                    'test_independent_topics_b',
                                    group='oslo_messaging_kafka')
         listener_b = self.useFixture(
-            utils.NotificationFixture(self.conf, self.url, ['2']))
+            utils.NotificationFixture(self.conf, self.notify_url, ['2']))
 
         a = listener_a.notifier('pub-1', topics=['1'])
         b = listener_b.notifier('pub-2', topics=['2'])
@@ -484,13 +488,13 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
 
     def test_all_categories(self):
         get_timeout = 1
-        if self.url.startswith("kafka://"):
+        if self.notify_url.startswith("kafka://"):
             get_timeout = 5
             self.conf.set_override('consumer_group', 'test_all_categories',
                                    group='oslo_messaging_kafka')
 
         listener = self.useFixture(utils.NotificationFixture(
-            self.conf, self.url, ['test_all_categories']))
+            self.conf, self.notify_url, ['test_all_categories']))
         n = listener.notifier('abc')
 
         cats = ['debug', 'audit', 'info', 'warn', 'error', 'critical']
@@ -513,26 +517,26 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
     def test_simple_batch(self):
         get_timeout = 3
         batch_timeout = 2
-        if self.url.startswith("amqp:"):
+        if self.notify_url.startswith("amqp:"):
             backend = os.environ.get("AMQP1_BACKEND")
             if backend == "qdrouterd":
                 # end-to-end acknowledgement with router intermediary
                 # sender pends until batch_size or timeout reached
                 self.skipTest("qdrouterd backend")
-        if self.url.startswith("kafka://"):
+        if self.notify_url.startswith("kafka://"):
             get_timeout = 10
             batch_timeout = 5
             self.conf.set_override('consumer_group', 'test_simple_batch',
                                    group='oslo_messaging_kafka')
 
         listener = self.useFixture(
-            utils.BatchNotificationFixture(self.conf, self.url,
+            utils.BatchNotificationFixture(self.conf, self.notify_url,
                                            ['test_simple_batch'],
                                            batch_size=100,
                                            batch_timeout=batch_timeout))
         notifier = listener.notifier('abc')
 
-        for i in six.moves.range(0, 205):
+        for i in range(0, 205):
             notifier.info({}, 'test%s' % i, 'Hello World!')
         events = listener.get_events(timeout=get_timeout)
         self.assertEqual(3, len(events))
@@ -542,10 +546,10 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
 
     def test_compression(self):
         get_timeout = 1
-        if self.url.startswith("amqp:"):
+        if self.notify_url.startswith("amqp:"):
             self.conf.set_override('kombu_compression', 'gzip',
                                    group='oslo_messaging_rabbit')
-        if self.url.startswith("kafka://"):
+        if self.notify_url.startswith("kafka://"):
             get_timeout = 5
             self.conf.set_override('compression_codec', 'gzip',
                                    group='oslo_messaging_kafka')
@@ -553,7 +557,7 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
                                    group='oslo_messaging_kafka')
 
         listener = self.useFixture(
-            utils.NotificationFixture(self.conf, self.url,
+            utils.NotificationFixture(self.conf, self.notify_url,
                                       ['test_compression']))
         notifier = listener.notifier('abc')
 
@@ -563,3 +567,35 @@ class NotifyTestCase(utils.SkipIfNoTransportURL):
         self.assertEqual('test', event[1])
         self.assertEqual('Hello World!', event[2])
         self.assertEqual('abc', event[3])
+
+
+class MetricsTestCase(utils.SkipIfNoTransportURL):
+
+    def setUp(self):
+        super().setUp(conf=cfg.ConfigOpts())
+        if self.rpc_url.startswith("kafka://"):
+            self.skipTest("kafka does not support RPC API")
+
+        self.config(metrics_enabled=True,
+                    group='oslo_messaging_metrics')
+
+    def test_functional(self):
+        # verify call metrics is sent and reflected in oslo.metrics
+        self.config(metrics_socket_file='/var/tmp/metrics_collector.sock',
+                    group='oslo_messaging_metrics')
+        metric_server = subprocess.Popen(["python3", "-m", "oslo_metrics"])
+        time.sleep(1)
+        group = self.useFixture(
+            utils.RpcServerGroupFixture(self.conf, self.rpc_url))
+        client = group.client(1)
+        client.add(increment=1)
+        time.sleep(1)
+        r = requests.get('http://localhost:3000', timeout=10)
+        for line in r.text.split('\n'):
+            if 'client_invocation_start_total{' in line:
+                self.assertEqual('1.0', line[-3:])
+            elif 'client_invocation_end_total{' in line:
+                self.assertEqual('1.0', line[-3:])
+            elif 'client_processing_seconds_count{' in line:
+                self.assertEqual('1.0', line[-3:])
+        metric_server.terminate()
