@@ -1,4 +1,3 @@
-
 # Copyright 2013 Red Hat, Inc.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
@@ -14,12 +13,13 @@
 #    under the License.
 
 import threading
-import warnings
+from unittest import mock
+
 
 import eventlet
 import fixtures
 from oslo_config import cfg
-from six.moves import mock
+from oslo_utils import eventletutils
 import testscenarios
 
 import oslo_messaging
@@ -32,12 +32,14 @@ from oslo_messaging.tests import utils as test_utils
 load_tests = testscenarios.load_tests_apply_scenarios
 
 
-class ServerSetupMixin(object):
+class ServerSetupMixin:
 
-    class Server(object):
-        def __init__(self, transport, topic, server, endpoint, serializer):
+    class Server:
+        def __init__(self, transport, topic, server, endpoint, serializer,
+                     exchange):
             self.controller = ServerSetupMixin.ServerController()
-            target = oslo_messaging.Target(topic=topic, server=server)
+            target = oslo_messaging.Target(topic=topic, server=server,
+                                           exchange=exchange)
             self.server = oslo_messaging.get_rpc_server(transport,
                                                         target,
                                                         [endpoint,
@@ -57,14 +59,14 @@ class ServerSetupMixin(object):
         def start(self):
             self.server.start()
 
-    class ServerController(object):
+    class ServerController:
         def __init__(self):
-            self.stopped = threading.Event()
+            self.stopped = eventletutils.Event()
 
         def stop(self, ctxt):
             self.stopped.set()
 
-    class TestSerializer(object):
+    class TestSerializer:
 
         def serialize_entity(self, ctxt, entity):
             return ('s' + entity) if entity else entity
@@ -73,82 +75,109 @@ class ServerSetupMixin(object):
             return ('d' + entity) if entity else entity
 
         def serialize_context(self, ctxt):
-            return dict([(k, 's' + v) for k, v in ctxt.items()])
+            return {k: 's' + v for k, v in ctxt.items()}
 
         def deserialize_context(self, ctxt):
-            return dict([(k, 'd' + v) for k, v in ctxt.items()])
+            return {k: 'd' + v for k, v in ctxt.items()}
 
     def __init__(self):
         self.serializer = self.TestSerializer()
 
-    def _setup_server(self, transport, endpoint, topic=None, server=None):
+    def _setup_server(self, transport, endpoint, topic=None, server=None,
+                      exchange=None):
         server = self.Server(transport,
                              topic=topic or 'testtopic',
                              server=server or 'testserver',
                              endpoint=endpoint,
-                             serializer=self.serializer)
+                             serializer=self.serializer,
+                             exchange=exchange)
 
         server.start()
         return server
 
-    def _stop_server(self, client, server, topic=None):
-        if topic is not None:
-            client = client.prepare(topic=topic)
+    def _stop_server(self, client, server, topic=None, exchange=None):
         client.cast({}, 'stop')
         server.wait()
 
-    def _setup_client(self, transport, topic='testtopic'):
-        return oslo_messaging.RPCClient(transport,
-                                        oslo_messaging.Target(topic=topic),
-                                        serializer=self.serializer)
+    def _setup_client(self, transport, topic='testtopic', exchange=None):
+        target = oslo_messaging.Target(topic=topic, exchange=exchange)
+        return oslo_messaging.get_rpc_client(transport, target=target,
+                                             serializer=self.serializer)
 
 
 class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
 
     def __init__(self, *args):
-        super(TestRPCServer, self).__init__(*args)
+        super().__init__(*args)
         ServerSetupMixin.__init__(self)
 
     def setUp(self):
-        super(TestRPCServer, self).setUp(conf=cfg.ConfigOpts())
+        super().setUp(conf=cfg.ConfigOpts())
+        # FakeExchangeManager uses a class-level exchanges mapping; "reset" it
+        # before tests assert amount of items stored
+        self.useFixture(fixtures.MonkeyPatch(
+            'oslo_messaging._drivers.impl_fake.FakeExchangeManager._exchanges',
+            new_value={}))
 
     def test_constructor(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
         target = oslo_messaging.Target(topic='foo', server='bar')
         endpoints = [object()]
         serializer = object()
         access_policy = dispatcher.DefaultRPCAccessPolicy
 
-        with warnings.catch_warnings(record=True) as capture:
-            warnings.simplefilter("always", FutureWarning)
-            server = oslo_messaging.get_rpc_server(transport,
-                                                   target,
-                                                   endpoints,
-                                                   serializer=serializer,
-                                                   access_policy=access_policy)
-        self.assertEqual(0, len(capture))
+        server = oslo_messaging.get_rpc_server(transport,
+                                               target,
+                                               endpoints,
+                                               serializer=serializer,
+                                               access_policy=access_policy,
+                                               executor='threading')
         self.assertIs(server.conf, self.conf)
         self.assertIs(server.transport, transport)
         self.assertIsInstance(server.dispatcher, oslo_messaging.RPCDispatcher)
         self.assertIs(server.dispatcher.endpoints, endpoints)
         self.assertIs(server.dispatcher.serializer, serializer)
-        self.assertEqual('blocking', server.executor_type)
+        self.assertEqual('threading', server.executor_type)
 
-    def test_constructor_without_explicit_RPCAccessPolicy(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+    def test_constructor_with_eventlet_executor(self):
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
         target = oslo_messaging.Target(topic='foo', server='bar')
         endpoints = [object()]
         serializer = object()
-        with warnings.catch_warnings(record=True) as capture:
-            warnings.simplefilter("always", FutureWarning)
-            oslo_messaging.get_rpc_server(transport, target,
-                                          endpoints, serializer=serializer)
-        self.assertEqual(1, len(capture))
-        w = capture[0]
-        self.assertEqual(FutureWarning, w.category)
+        access_policy = dispatcher.DefaultRPCAccessPolicy
+
+        server = oslo_messaging.get_rpc_server(transport,
+                                               target,
+                                               endpoints,
+                                               serializer=serializer,
+                                               access_policy=access_policy,
+                                               executor='eventlet')
+        self.assertIs(server.conf, self.conf)
+        self.assertIs(server.transport, transport)
+        self.assertIsInstance(server.dispatcher, oslo_messaging.RPCDispatcher)
+        self.assertIs(server.dispatcher.endpoints, endpoints)
+        self.assertIs(server.dispatcher.serializer, serializer)
+        self.assertEqual('eventlet', server.executor_type)
+
+    def test_constructor_with_unrecognized_executor(self):
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
+        target = oslo_messaging.Target(topic='foo', server='bar')
+        endpoints = [object()]
+        serializer = object()
+        access_policy = dispatcher.DefaultRPCAccessPolicy
+
+        self.assertRaises(
+            server_module.ExecutorLoadFailure,
+            oslo_messaging.get_rpc_server,
+            transport=transport,
+            target=target,
+            endpoints=endpoints,
+            serializer=serializer,
+            access_policy=access_policy,
+            executor='boom')
 
     def test_server_wait_method(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
         target = oslo_messaging.Target(topic='foo', server='bar')
         endpoints = [object()]
         serializer = object()
@@ -161,7 +190,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
             """
 
             def __init__(self, *args, **kwargs):
-                super(MagicMockIgnoreArgs, self).__init__()
+                super().__init__()
 
         server = oslo_messaging.get_rpc_server(transport, target, endpoints,
                                                serializer=serializer)
@@ -180,7 +209,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self.assertEqual(1, listener.cleanup.call_count)
 
     def test_no_target_server(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
         server = oslo_messaging.get_rpc_server(
             transport,
@@ -195,7 +224,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
             self.assertTrue(False)
 
     def test_no_server_topic(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
         target = oslo_messaging.Target(server='testserver')
         server = oslo_messaging.get_rpc_server(transport, target, [])
         try:
@@ -207,7 +236,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
             self.assertTrue(False)
 
     def _test_no_client_topic(self, call=True):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
         client = self._setup_client(transport, topic=None)
 
@@ -228,12 +257,12 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self._test_no_client_topic(call=False)
 
     def test_client_call_timeout(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
         finished = False
         wait = threading.Condition()
 
-        class TestEndpoint(object):
+        class TestEndpoint:
             def ping(self, ctxt, arg):
                 with wait:
                     if not finished:
@@ -256,7 +285,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self._stop_server(client, server_thread)
 
     def test_unknown_executor(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
         try:
             oslo_messaging.get_rpc_server(transport, None, [], executor='foo')
@@ -267,9 +296,9 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
             self.assertTrue(False)
 
     def test_cast(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
-        class TestEndpoint(object):
+        class TestEndpoint:
             def __init__(self):
                 self.pings = []
 
@@ -288,14 +317,20 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self.assertEqual(['dsfoo', 'dsbar'], endpoint.pings)
 
     def test_call(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        # NOTE(milan): using a separate transport instance for each the client
+        # and the server to be able to check independent transport instances
+        # can communicate over same exchange&topic
+        transport_srv = oslo_messaging.get_rpc_transport(self.conf,
+                                                         url='fake:')
+        transport_cli = oslo_messaging.get_rpc_transport(self.conf,
+                                                         url='fake:')
 
-        class TestEndpoint(object):
+        class TestEndpoint:
             def ping(self, ctxt, arg):
                 return arg
 
-        server_thread = self._setup_server(transport, TestEndpoint())
-        client = self._setup_client(transport)
+        server_thread = self._setup_server(transport_srv, TestEndpoint())
+        client = self._setup_client(transport_cli)
 
         self.assertIsNone(client.call({}, 'ping', arg=None))
         self.assertEqual(0, client.call({}, 'ping', arg=0))
@@ -307,9 +342,9 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self._stop_server(client, server_thread)
 
     def test_direct_call(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
-        class TestEndpoint(object):
+        class TestEndpoint:
             def ping(self, ctxt, arg):
                 return arg
 
@@ -327,9 +362,9 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self._stop_server(client, server_thread)
 
     def test_context(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
-        class TestEndpoint(object):
+        class TestEndpoint:
             def ctxt_check(self, ctxt, key):
                 return ctxt[key]
 
@@ -344,9 +379,9 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self._stop_server(client, server_thread)
 
     def test_failure(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
-        class TestEndpoint(object):
+        class TestEndpoint:
             def ping(self, ctxt, arg):
                 raise ValueError(arg)
 
@@ -376,7 +411,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         except Exception as ex:
             self.assertIsInstance(ex, ValueError)
             self.assertEqual('dsfoo', str(ex))
-            self.assertTrue(len(debugs) == 0)
+            self.assertTrue(len(debugs) == 2)
             self.assertGreater(len(errors), 0)
         else:
             self.assertTrue(False)
@@ -384,7 +419,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self._stop_server(client, server_thread)
 
     def test_expected_failure(self):
-        transport = oslo_messaging.get_transport(self.conf, url='fake:')
+        transport = oslo_messaging.get_rpc_transport(self.conf, url='fake:')
 
         debugs = []
         errors = []
@@ -404,7 +439,7 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
         self.useFixture(fixtures.MockPatchObject(
             rpc_server_module.LOG, 'error', stub_error))
 
-        class TestEndpoint(object):
+        class TestEndpoint:
             @oslo_messaging.expected_exceptions(ValueError)
             def ping(self, ctxt, arg):
                 raise ValueError(arg)
@@ -424,6 +459,20 @@ class TestRPCServer(test_utils.BaseTestCase, ServerSetupMixin):
 
         self._stop_server(client, server_thread)
 
+    @mock.patch('oslo_messaging.rpc.server.LOG')
+    def test_warning_when_notifier_transport(self, log):
+        transport = oslo_messaging.get_notification_transport(self.conf)
+        target = oslo_messaging.Target(topic='foo', server='bar')
+        endpoints = [object()]
+        serializer = object()
+
+        oslo_messaging.get_rpc_server(transport, target,
+                                      endpoints, serializer=serializer)
+        log.warning.assert_called_once_with(
+            "Using notification transport for RPC. Please use "
+            "get_rpc_transport to obtain an RPC transport "
+            "instance.")
+
 
 class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
 
@@ -486,8 +535,8 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
             single_server = params['server1'] == params['server2']
             return not (single_topic and single_server)
 
-        # fanout to multiple servers on same topic and exchange
-        # each endpoint will receive both messages
+        # fanout to multiple servers on same topic and exchange each endpoint
+        # will receive both messages
         def fanout_to_servers(scenario):
             params = scenario[1]
             fanout = params['fanout1'] or params['fanout2']
@@ -519,23 +568,25 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
             cls.scenarios = [m(i) for i in cls.scenarios]
 
     def __init__(self, *args):
-        super(TestMultipleServers, self).__init__(*args)
+        super().__init__(*args)
         ServerSetupMixin.__init__(self)
 
     def setUp(self):
-        super(TestMultipleServers, self).setUp(conf=cfg.ConfigOpts())
+        super().setUp(conf=cfg.ConfigOpts())
+        self.useFixture(fixtures.MonkeyPatch(
+            'oslo_messaging._drivers.impl_fake.FakeExchangeManager._exchanges',
+            new_value={}))
 
     def test_multiple_servers(self):
-        url1 = 'fake:///' + (self.exchange1 or '')
-        url2 = 'fake:///' + (self.exchange2 or '')
-
-        transport1 = oslo_messaging.get_transport(self.conf, url=url1)
-        if url1 != url2:
-            transport2 = oslo_messaging.get_transport(self.conf, url=url1)
+        transport1 = oslo_messaging.get_rpc_transport(self.conf,
+                                                      url='fake:')
+        if self.exchange1 != self.exchange2:
+            transport2 = oslo_messaging.get_rpc_transport(self.conf,
+                                                          url='fake:')
         else:
             transport2 = transport1
 
-        class TestEndpoint(object):
+        class TestEndpoint:
             def __init__(self):
                 self.pings = []
 
@@ -551,12 +602,18 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
             endpoint1 = endpoint2 = TestEndpoint()
 
         server1 = self._setup_server(transport1, endpoint1,
-                                     topic=self.topic1, server=self.server1)
+                                     topic=self.topic1,
+                                     exchange=self.exchange1,
+                                     server=self.server1)
         server2 = self._setup_server(transport2, endpoint2,
-                                     topic=self.topic2, server=self.server2)
+                                     topic=self.topic2,
+                                     exchange=self.exchange2,
+                                     server=self.server2)
 
-        client1 = self._setup_client(transport1, topic=self.topic1)
-        client2 = self._setup_client(transport2, topic=self.topic2)
+        client1 = self._setup_client(transport1, topic=self.topic1,
+                                     exchange=self.exchange1)
+        client2 = self._setup_client(transport2, topic=self.topic2,
+                                     exchange=self.exchange2)
 
         client1 = client1.prepare(server=self.server1)
         client2 = client2.prepare(server=self.server2)
@@ -572,9 +629,9 @@ class TestMultipleServers(test_utils.BaseTestCase, ServerSetupMixin):
         (client2.call if self.call2 else client2.cast)({}, 'ping', arg='2')
 
         self._stop_server(client1.prepare(fanout=None),
-                          server1, topic=self.topic1)
+                          server1, topic=self.topic1, exchange=self.exchange1)
         self._stop_server(client2.prepare(fanout=None),
-                          server2, topic=self.topic2)
+                          server2, topic=self.topic2, exchange=self.exchange2)
 
         def check(pings, expect):
             self.assertEqual(len(expect), len(pings))
@@ -593,7 +650,7 @@ TestMultipleServers.generate_scenarios()
 
 class TestServerLocking(test_utils.BaseTestCase):
     def setUp(self):
-        super(TestServerLocking, self).setUp(conf=cfg.ConfigOpts())
+        super().setUp(conf=cfg.ConfigOpts())
 
         def _logmethod(name):
             def method(self, *args, **kwargs):
@@ -603,7 +660,7 @@ class TestServerLocking(test_utils.BaseTestCase):
 
         executors = []
 
-        class FakeExecutor(object):
+        class FakeExecutor:
             def __init__(self, *args, **kwargs):
                 self._lock = threading.Lock()
                 self._calls = []
@@ -658,11 +715,11 @@ class TestServerLocking(test_utils.BaseTestCase):
         # Test that if 2 threads call a method simultaneously, both will wait,
         # but only 1 will call the underlying executor method.
 
-        start_event = threading.Event()
-        finish_event = threading.Event()
+        start_event = eventletutils.Event()
+        finish_event = eventletutils.Event()
 
-        running_event = threading.Event()
-        done_event = threading.Event()
+        running_event = eventletutils.Event()
+        done_event = eventletutils.Event()
 
         _runner = [None]
 
@@ -673,7 +730,7 @@ class TestServerLocking(test_utils.BaseTestCase):
                 running_event.set()
 
                 start_event.wait()
-                super(SteppingFakeExecutor, self).__init__(*args, **kwargs)
+                super().__init__(*args, **kwargs)
                 done_event.set()
 
                 finish_event.wait()
@@ -688,7 +745,7 @@ class TestServerLocking(test_utils.BaseTestCase):
         runner = _runner[0]
         waiter = start2 if runner == start1 else start2
 
-        waiter_finished = threading.Event()
+        waiter_finished = eventletutils.Event()
         waiter.link(lambda _: waiter_finished.set())
 
         # At this point, runner is running start(), and waiter() is waiting for
@@ -737,8 +794,8 @@ class TestServerLocking(test_utils.BaseTestCase):
         # Ensure that if 2 threads wait for the completion of 'start', the
         # first will wait until complete_event is signalled, but the second
         # will continue
-        complete_event = threading.Event()
-        complete_waiting_callback = threading.Event()
+        complete_event = eventletutils.Event()
+        complete_waiting_callback = eventletutils.Event()
 
         start_state = self.server._states['start']
         old_wait_for_completion = start_state.wait_for_completion
@@ -755,7 +812,7 @@ class TestServerLocking(test_utils.BaseTestCase):
 
         # thread1 will wait for start to complete until we signal it
         thread1 = eventlet.spawn(self.server.stop)
-        thread1_finished = threading.Event()
+        thread1_finished = eventletutils.Event()
         thread1.link(lambda _: thread1_finished.set())
 
         self.server.start()
@@ -801,7 +858,7 @@ class TestServerLocking(test_utils.BaseTestCase):
         # Test that we generate a log message if we wait longer than
         # DEFAULT_LOG_AFTER
 
-        log_event = threading.Event()
+        log_event = eventletutils.Event()
         mock_log.warning.side_effect = lambda _, __: log_event.set()
 
         # Call stop without calling start. We should log a wait after 1 second
@@ -817,7 +874,7 @@ class TestServerLocking(test_utils.BaseTestCase):
         # Test that we generate a log message if we wait longer than
         # the number of seconds passed to log_after
 
-        log_event = threading.Event()
+        log_event = eventletutils.Event()
         mock_log.warning.side_effect = lambda _, __: log_event.set()
 
         # Call stop without calling start. We should log a wait after 1 second
@@ -833,7 +890,7 @@ class TestServerLocking(test_utils.BaseTestCase):
         # Test that we log a message after log_after seconds if we've also
         # specified an absolute timeout
 
-        log_event = threading.Event()
+        log_event = eventletutils.Event()
         mock_log.warning.side_effect = lambda _, __: log_event.set()
 
         # Call stop without calling start. We should log a wait after 1 second
@@ -858,7 +915,7 @@ class TestServerLocking(test_utils.BaseTestCase):
         # Start the server, which will also instantiate an executor
         self.server.start()
         self.server.stop()
-        shutdown_called = threading.Event()
+        shutdown_called = eventletutils.Event()
 
         # Patch the executor's stop method to be very slow
         def slow_shutdown(wait):
